# -*- coding: utf-8 -*-
'''
Database object store.

The shove URI used for database object stores is the format used by
SQLAlchemy:

<driver>://<username>:<password>@<host>:<port>/<database>

<driver> is the database engine. The engines currently supported SQLAlchemy are
sqlite, mysql, postgres, oracle, mssql, and firebird.
<username> is the database account user name
<password> is the database accound password
<host> is the database location
<port> is the database port
<database> is the name of the specific database

For more information on specific databases see:

http://www.sqlalchemy.org/docs/dbengine.myt#dbengine_supported
'''

from stuf.six import native
try:
    from sqlalchemy import LargeBinary as Binary
except ImportError:
    from sqlalchemy import Binary
try:
    from sqlalchemy import MetaData, Table, Column, String, select
except ImportError:
    raise ImportError('requires SQLAlchemy >= 0.4')

from shove.store import BaseStore

__all__ = ['DBStore']


class DBStore(BaseStore):

    '''
    Database cache backend.
    '''

    def __init__(self, engine, **kw):
        super(DBStore, self).__init__(engine, **kw)
        # make store table
        self._store = Table(
            # get tablename
            kw.get('tablename', 'store'),
            MetaData(engine),
            Column('key', String(255), primary_key=True, nullable=False),
            Column('value', Binary, nullable=False),
        )
        # create store table if it does not exist
        if not self._store.exists():
            self._store.create()

    def __getitem__(self, key):
        row = select(
            [self._store.c.value], self._store.c.key == key,
        ).execute().fetchone()
        if row is not None:
            return self.loads(native(row.value))
        raise KeyError(key)

    def __setitem__(self, k, v):
        v, store = self.dumps(v), self._store
        # update database if key already present
        if k in self:
            store.update(store.c.key == k).execute(value=v)
        # insert new key if key not present
        else:
            store.insert().execute(key=k, value=v)

    def __delitem__(self, key):
        self._store.delete(self._store.c.key == key).execute()

    def __iter__(self):
        for item in select([self._store.c.key]).execute().fetchall():
            yield item[0]

    def __len__(self):
        return self._store.count().execute().fetchone()[0]

    def clear(self):
        self._store.delete().execute()



# -*- coding: utf-8 -*-
'''
git-based object store.

shove's URI for git-based stores follows the form:

git://<path>

Where the path is a URI path to a git repository on a local filesystem.
Alternatively, a native pathname to the repository can be passed as the
'engine' argument.
'''

try:
    from dulwich.repo import Repo
    from dulwich.errors import NotGitRepository
except ImportError:
    raise ImportError('requires dulwich library')

from shove.store import FileStore
from shove._compat import quote_plus


class GitStore(FileStore):

    init = 'git://'

    def __init__(self, engine, **kw):
        super(GitStore, self).__init__(engine, **kw)
        try:
            self._repo = Repo(self._dir)
        except NotGitRepository:
            self._repo = Repo.init(self._dir)

    def __setitem__(self, key, value):
        super(GitStore, self).__setitem__(key, value)
        fname = quote_plus(key)
        self._repo.stage([fname])
        self._repo.do_commit('added {0}'.format(fname), committer='shove')

    def __delitem__(self, key):
        super(GitStore, self).__delitem__(key)
        fname = quote_plus(key)
        self._repo.stage([fname])
        self._repo.do_commit('removed {0}'.format(fname))



# -*- coding: utf-8 -*-
'''shove core.'''

from os.path import exists, join
from os import listdir, remove, makedirs
from zlib import compress, decompress, error

from stuf.utils import ld, optimize
from stuf.six import HIGHEST_PROTOCOL

from shove._compat import url2pathname, quote_plus, unquote_plus


class Base(object):

    '''
    Base.
    '''

    def __init__(self, engine, **kw):
        # keyword compress True, False, or an integer compression level (1-9)
        self._compress = kw.get('compress', False)
        # pickle protocol
        self._protocol = kw.get('protocol', HIGHEST_PROTOCOL)

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def dumps(self, value):
        '''
        Optionally serializes and compresses an object.
        '''
        # serialize anything but ASCII strings
        value = optimize(value)
        compression = self._compress
        if compression:
            value = compress(value, 9 if compression else compression)
        return value

    def loads(self, value):
        '''
        Deserializes and optionally decompresses an object.
        '''
        if self._compress:
            try:
                value = decompress(value)
            except error:
                pass
        return ld(value)


class Mapping(Base):

    '''
    Base mapping.
    '''

    def __getitem__(self, key):
        try:
            return self._store[key]
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        try:
            del self._store[key]
        except:
            raise KeyError(key)

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def dumps(self, value):
        '''
        Optionally serializes and compresses an object.
        '''
        # serialize anything but ASCII strings
        value = optimize(value)
        compression = self._compress
        if compression:
            value = compress(value, 9 if compression else compression)
        return value

    def loads(self, value):
        '''
        Deserializes and optionally decompresses an object.
        '''
        if self._compress:
            try:
                value = decompress(value)
            except error:
                pass
        return ld(value)


class FileBase(Base):

    '''
    Base for file based storage.
    '''

    def __init__(self, engine, **kw):
        super(FileBase, self).__init__(engine, **kw)
        if engine.startswith(self.init):
            engine = url2pathname(engine.split('://')[1])
        self._dir = engine
        # Create directory
        if not exists(self._dir):
            self._createdir()

    def __getitem__(self, key):
        # (per Larry Meyn)
        try:
            with open(self._key_to_file(key), 'rb') as item:
                return self.loads(item.read())
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        # (per Larry Meyn)
        try:
            with open(self._key_to_file(key), 'wb') as item:
                item.write(self.dumps(value))
        except (IOError, OSError):
            raise KeyError(key)

    def __delitem__(self, key):
        try:
            remove(self._key_to_file(key))
        except (IOError, OSError):
            raise KeyError(key)

    def __iter__(self):
        for name in listdir(self._dir):
            if not name.startswith('.'):
                yield unquote_plus(name)

    def __contains__(self, key):
        return exists(self._key_to_file(key))

    def __len__(self):
        return len([i for i in listdir(self._dir) if not i.startswith('.')])

    def _createdir(self):
        # creates the store directory
        try:
            makedirs(self._dir)
        except OSError:
            raise EnvironmentError(
                'Cache directory "{0}" does not exist and could not be '
                'created'.format(self._dir)
            )

    def _key_to_file(self, key):
        # gives the filesystem path for a key
        return join(self._dir, quote_plus(key))



# -*- coding: utf-8 -*-
'''shove multishove tests'''

from stuf.six import unittest


def setUpModule():
    import os
    from tempfile import mkdtemp
    TMP = mkdtemp()
    os.chdir(TMP)


class Multi(object):

    def test__getitem__(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store['max'], 3)

    def test__setitem__(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store['max'], 3)

    def test__delitem__(self):
        self.store['max'] = 3
        self.store.sync()
        del self.store['max']
        self.store.sync()
        self.assertEqual('max' in self.store, False)

    def test_get(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store.get('min'), None)

    def test__cmp__(self):
        from shove.core import MultiShove
        tstore = MultiShove()
        self.store['max'] = 3
        tstore['max'] = 3
        self.store.sync()
        tstore.sync()
        self.assertEqual(self.store, tstore)

    def test__len__(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        self.assertEqual(len(self.store), 3)

    def test_clear(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.clear()
        self.assertEqual(len(self.store), 0)

    def test_items(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(self.store.items())
        self.assertEqual(('min', 6) in slist, True)

    def test_keys(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(self.store.keys())
        self.assertEqual('min' in slist, True)

    def test_values(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(self.store.values())
        self.assertEqual(6 in slist, True)

    def test_pop(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store.sync()
        item = self.store.pop('min')
        self.store.sync()
        self.assertEqual(item, 6)

    def test_popitem(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        item = self.store.popitem()
        self.store.sync()
        self.assertEqual(len(item) + len(self.store), 4)

    def test_setdefault(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['powl'] = 7
        self.store.sync()
        self.store.setdefault('pow', 8)
        self.store.sync()
        self.assertEqual(self.store['pow'], 8)

    def test_update(self):
        from shove.core import MultiShove
        tstore = MultiShove()
        tstore['max'] = 3
        tstore['min'] = 6
        tstore['pow'] = 7
        self.store['max'] = 2
        self.store['min'] = 3
        self.store['pow'] = 7
        self.store.update(tstore)
        self.store.sync()
        self.assertEqual(self.store['min'], 6)


class TestMultiShove(Multi, unittest.TestCase):

    stores = (
        'simple://', 'dbm://one.dbm', 'memory://', 'file://two', 'sqlite://',
    )

    def setUp(self):
        from shove.core import MultiShove
        self.store = MultiShove(*self.stores)

    def tearDown(self):
        import os
        import shutil
        self.store.close()
        shutil.rmtree('two')
        try:
            os.remove('one.dbm')
        except OSError:
            os.remove('one.dbm.db')


class TestThreadShove(unittest.TestCase):

    stores = (
        'simple://', 'memory://', 'file://six', 'sqlite://',
    )

    def setUp(self):
        from shove.core import ThreadShove
        self.store = ThreadShove(*self.stores, max_workers=2)

    def tearDown(self):
        import shutil
        self.store.close()
        shutil.rmtree('six')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''shove store tests'''

from stuf.six import PY3, unittest, keys, values, items

TMP = None


def setUpModule():
    import os
    from tempfile import mkdtemp
    TMP = mkdtemp()
    os.chdir(TMP)


class EventualStore(object):

    def setUp(self):
        from shove import Shove
        self.store = Shove(self.initstring, compress=True)

    def tearDown(self):
        self.store.close()

    def test__getitem__(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store['max'], 3)

    def test__setitem__(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store['max'], 3)

    def test__delitem__(self):
        self.store['max'] = 3
        self.store.sync()
        del self.store['max']
        self.store.sync()
        self.assertEqual('max' in self.store, False)

    def test_get(self):
        self.store['max'] = 3
        self.store.sync()
        self.assertEqual(self.store.get('min'), None)

    def test__cmp__(self):
        from shove import Shove
        tstore = Shove()
        self.store['max'] = 3
        self.store.sync()
        tstore['max'] = 3
        self.assertEqual(self.store, tstore)

    def test__len__(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store.sync()
        self.assertEqual(len(self.store), 2)

    def test_items(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(items(self.store))
        self.assertEqual(('min', 6) in slist, True)

    def test_keys(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(keys(self.store))
        self.assertEqual('min' in slist, True)

    def test_values(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        slist = list(values(self.store))
        self.assertEqual(6 in slist, True)

    def test_pop(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store.sync()
        item = self.store.pop('min')
        self.store.sync()
        self.assertEqual(item, 6)

    def test_setdefault(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store.sync()
        self.assertEqual(self.store.setdefault('pow', 8), 8)
        self.store.sync()
        self.assertEqual(self.store['pow'], 8)

    def test_update(self):
        from shove import Shove
        tstore = Shove()
        tstore['max'] = 3
        tstore['min'] = 6
        tstore['pow'] = 7
        self.store['max'] = 2
        self.store['min'] = 3
        self.store['pow'] = 7
        self.store.update(tstore)
        self.store.sync()
        self.assertEqual(self.store['min'], 6)


class Store(EventualStore):

    def test_clear(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        self.store.clear()
        self.store.sync()
        self.assertEqual(len(self.store), 0)

    def test_popitem(self):
        self.store['max'] = 3
        self.store['min'] = 6
        self.store['pow'] = 7
        self.store.sync()
        item = self.store.popitem()
        self.store.sync()
        self.assertEqual(len(item) + len(self.store), 4)

    def test_close(self):
        self.store.close()
        self.assertEqual(self.store._store, None)
        self.assertEqual(self.store._buffer, None)
        self.assertEqual(self.store._cache, None)


class TestSimpleStore(Store, unittest.TestCase):

    initstring = 'simple://'


class TestMemoryStore(Store, unittest.TestCase):

    initstring = 'memory://'


class TestFileStore(Store, unittest.TestCase):

    initstring = 'file://test'

    def tearDown(self):
        import shutil
        self.store.close()
        shutil.rmtree('test')


class TestHgStore(Store, unittest.TestCase):

    initstring = 'hg://test3'

    def tearDown(self):
        import shutil
        self.store.close()
        shutil.rmtree('test3')


class TestGitStore(Store, unittest.TestCase):

    initstring = 'git://test4'

    def tearDown(self):
        import shutil
        self.store.close()
        shutil.rmtree('test4')


class TestDBMStore(Store, unittest.TestCase):

    initstring = 'dbm://test.dbm'

    def tearDown(self):
        import os
        self.store.close()
        try:
            os.remove('test.dbm')
        except OSError:
            os.remove('test.dbm.db')


class TestDBStore(Store, unittest.TestCase):

    initstring = 'sqlite://'


class TestMongoDBStore(Store, unittest.TestCase):

    initstring = 'mongodb://127.0.0.1:27017/shove/shove'

    @classmethod
    def setUpClass(cls):
        import os
        from fabric.api import local
        os.mkdir('mongo')
        local('touch ./mongo/mongo.log')
        local('mongod --dbpath {0} --fork --logpath {1}'.format(
            './mongo/', './mongo/mongolog',
        ))

    def tearDown(self):
        self.store.clear()

    def test_close(self):
        pass

    @classmethod
    def tearDownClass(cls):
        from fabric.api import local
        local('killall mongod')


@unittest.skip('reason')
class TestFTPStore(Store, unittest.TestCase):

    initstring = 'put ftp string here'

    def setUp(self):
        from shove import Shove
        self.store = Shove(self.ftpstring, compress=True)

    def tearDown(self):
        self.store.clear()
        self.store.close()


@unittest.skip('reason')
class TestS3Store(Store, unittest.TestCase):

    initstring = 's3 test string here'

    def tearDown(self):
        self.store.clear()
        self.store.close()


if not PY3:
    class TestZODBStore(Store, unittest.TestCase):

        initstring = 'zodb://test.db'

        def tearDown(self):
            self.store.close()
            import os
            os.remove('test.db')
            os.remove('test.db.index')
            os.remove('test.db.tmp')
            os.remove('test.db.lock')

    class TestDurusStore(Store, unittest.TestCase):

        initstring = 'durus://test.durus'

        def tearDown(self):
            import os
            self.store.close()
            os.remove('test.durus')

    class TestRedisStore(Store, unittest.TestCase):

        initstring = 'redis://localhost:6379/0'

        @classmethod
        def setUpClass(cls):
            from fabric.api import local
            local('redis-server &')

        def tearDown(self):
            if self.store._store is not None:
                self.store.clear()
                self.store.close()

        @classmethod
        def tearDownClass(cls):
            from fabric.api import local
            local('killall redis-server')

    @unittest.skip('reason')
    class TestBSDBStore(Store, unittest.TestCase):

        initstring = 'bsddb://test.db'

        def tearDown(self):
            import os
            self.store.close()
            os.remove('test.db')

    class TestCassandraStore(EventualStore, unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            from fabric.api import local
            local('cassandra -f &')

        def setUp(self):
            from shove import Shove
            from pycassa.system_manager import SystemManager  # @UnresolvedImport @IgnorePep8
            system_manager = SystemManager('localhost:9160')
            try:
                system_manager.create_column_family('Murk', 'shove')
            except:
                pass
            self.store = Shove('cassandra://localhost:9160/Murk/shove')

        def tearDown(self):
            self.store.clear()
            self.store.close()
            from pycassa.system_manager import SystemManager  # @UnresolvedImport @IgnorePep8
            system_manager = SystemManager('localhost:9160')
            system_manager.drop_column_family('Murk', 'shove')

        @classmethod
        def tearDownClass(cls):
            from fabric.api import local
            local('killall java')

    class TestHDF5Store(Store, unittest.TestCase):

        initstring = 'hdf5://test.hdf5/test'

        def setUp(self):
            from shove import Shove
            self.store = Shove()

        def tearDown(self):
            import os
            self.store.close()
            try:
                os.remove('test.hdf5')
            except OSError:
                pass

    @unittest.skip('reason')
    class TestLevelDBStore(unittest.TestCase):

        initstring = 'leveldb://test'

        def tearDown(self):
            import shutil
            shutil.rmtree('test')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''shove store support.'''

import shutil
from copy import deepcopy
from threading import Condition
from collections import MutableMapping

from shove.base import Mapping, FileBase
from shove._compat import anydbm, synchronized, url2pathname

__all__ = ('DBMStore', 'FileStore', 'MemoryStore', 'SimpleStore')


class BaseStore(Mapping, MutableMapping):

    '''Base store.'''

    def close(self):
        '''
        Closes internal store and clears object references.
        '''
        try:
            self._store.close()
        except AttributeError:
            pass
        self._store = None


class SimpleStore(BaseStore):

    '''
    Single-process in-memory store.

    The shove URI for a simple store is:

    simple://
    '''

    def __init__(self, engine, **kw):
        super(SimpleStore, self).__init__(engine, **kw)
        self._store = dict()


class MemoryStore(SimpleStore):

    '''
    Thread-safe in-memory store.

    The shove URI for a memory store is:

    memory://
    '''

    def __init__(self, engine, **kw):
        super(MemoryStore, self).__init__(engine, **kw)
        self._lock = Condition()

    @synchronized
    def __getitem__(self, key):
        return deepcopy(super(MemoryStore, self).__getitem__(key))

    __setitem__ = synchronized(SimpleStore.__setitem__)
    __delitem__ = synchronized(SimpleStore.__delitem__)


class ClientStore(BaseStore):

    '''
    Base store where updates must be committed to disk.
    '''

    def __init__(self, engine, **kw):
        super(ClientStore, self).__init__(engine, **kw)
        if engine.startswith(self.init):
            self._engine = url2pathname(engine.split('://')[1])

    def __getitem__(self, key):
        return self.loads(super(ClientStore, self).__getitem__(key))

    def __setitem__(self, key, value):
        super(ClientStore, self).__setitem__(key, self.dumps(value))


class SyncStore(ClientStore):

    '''
    Base store where updates have to be synced to disk.
    '''

    def __setitem__(self, key, value):
        super(SyncStore, self).__setitem__(key, value)
        try:
            self.sync()
        except AttributeError:
            pass

    def __delitem__(self, key):
        super(SyncStore, self).__delitem__(key)
        try:
            self.sync()
        except AttributeError:
            pass


class DBMStore(SyncStore):

    '''
    DBM Database Store.

    shove's URI for DBM stores follows the form:

    dbm://<path>

    Where <path> is a URL path to a DBM database. Alternatively, the native
    pathname to a DBM database can be passed as the 'engine' parameter.
    '''

    init = 'dbm://'

    def __init__(self, engine, **kw):
        super(DBMStore, self).__init__(engine, **kw)
        self._store = anydbm.open(self._engine, 'c')
        try:
            self.sync = self._store.sync
        except AttributeError:
            pass

    def __iter__(self):
        return iter(self._store.keys())


class FileStore(FileBase, BaseStore):

    '''
    Filesystem-based object store.

    shove's URI for filesystem-based stores follows the form:

    file://<path>

    Where the path is a URI path to a directory on a local filesystem.
    Alternatively, a native pathname to the directory can be passed as the
    'engine' argument.
    '''

    init = 'file://'

    def clear(self):
        shutil.rmtree(self._dir)
        self._createdir()



# -*- coding: utf-8 -*-
'''
HDF5 Database Store.

shove's URI for HDF5 stores follows the form:

hdf5://<path>/<group>

Where <path> is a URL path to a HDF5 database. Alternatively, the native
pathname to a HDF5 database can be passed as the 'engine' parameter.
<group> is the name of the database.
'''

try:
    import h5py
except ImportError:
    raise ImportError('requires h5py library')

from shove.store import ClientStore

__all__ = ['HDF5Store']


class HDF5Store(ClientStore):

    '''LevelDB based store'''

    init = 'hdf5://'

    def __init__(self, engine, **kw):
        super(HDF5Store, self).__init__(engine, **kw)
        engine, group = self._engine.rsplit('/')
        self._store = h5py.File(engine).require_group(group).attrs



# -*- coding: utf-8 -*-
'''
Redis-based object cache

The shove URI for a redis cache is:

redis://<host>:<port>/<db>
'''

try:
    import redis
except ImportError:
    raise ImportError('requires redis')

from shove.base import Base
from shove._compat import urlsplit

__all__ = ['RedisCache']


class RedisCache(Base):

    '''Redis cache backend'''

    init = 'redis://'

    def __init__(self, engine, **kw):
        super(RedisCache, self).__init__(engine, **kw)
        spliturl = urlsplit(engine)
        host, port = spliturl[1].split(':')
        db = spliturl[2].replace('/', '')
        self._store = redis.Redis(host, int(port), db)
        # Set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        return self.loads(self._store[key])

    def __setitem__(self, key, value):
        self._store.setex(key, self.dumps(value), self.timeout)

    def __delitem__(self, key):
        self._store.delete(key)



# -*- coding: utf-8 -*-
'''
Mercurial-based object store.

shove's URI for Mercurial-based stores follows the form:

hg://<path>

Where the path is a URI path to a Mercurial repository on a local filesystem.
Alternatively, a native pathname to the repository can be passed as the
'engine' argument.
'''

try:
    import hgapi
except ImportError:
    raise ImportError('requires hgapi library')

from shove.store import FileStore


class HgStore(FileStore):

    init = 'hg://'

    def __init__(self, engine, **kw):
        super(HgStore, self).__init__(engine, **kw)
        self._repo = hgapi.Repo(self._dir)
        try:
            self._repo.hg_status()
        except Exception:
            self._repo.hg_init()

    def __setitem__(self, key, value):
        super(HgStore, self).__setitem__(key, value)
        fname = self._key_to_file(key)
        status = self._repo.hg_status()
        if fname in status['!']:
            self._repo.hg_add(fname)
            self._repo.hg_commit('added {0}'.format(fname))
        if fname in status['M']:
            self._repo.hg_commit('added {0}'.format(fname))

    def __delitem__(self, key):
        super(HgStore, self).__delitem__(key)
        fname = self._key_to_file(key)
        if fname in self._repo.hg_status()['R']:
            self._repo.hg_remove(fname)
            self._repo.hg_commit('removed {0}'.format(fname))



# -*- coding: utf-8 -*-
'''
"memcached" cache.

The shove URI for a memcache cache is:

memcache://<memcache_server>
'''

try:
    import memcache
except ImportError:
    raise ImportError("requires 'python-memcached' library")

from shove.base import Base

__all__ = ['MemCache']


class MemCache(Base):

    '''Memcached cache backend'''

    def __init__(self, engine, **kw):
        super(MemCache, self).__init__(engine, **kw)
        if engine.startswith('memcache://'):
            engine = engine.split('://')[1]
        self._store = memcache.Client(engine.split(';'))
        # set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        value = self._store.get(key)
        if value is None:
            raise KeyError(key)
        return self.loads(value)

    def __setitem__(self, key, value):
        self._store.set(key, self.dumps(value), self.timeout)

    def __delitem__(self, key):
        self._store.delete(key)



# -*- coding: utf-8 -*-
'''
MongoDB Database Store.

shove's URI for MongoDB stores follows the form:

mongodb://<host>:<port>/<db>/collection/
'''

try:
    from bson.binary import Binary
    from pymongo.connection import Connection
except ImportError:
    raise ImportError('requires pymongo library')

from shove._compat import urlsplit
from shove.store import SimpleStore

__all__ = ['MongoDBStore']


class MongoDBStore(SimpleStore):

    '''
    MongoDB based store.
    '''

    init = 'mongodb://'

    def __init__(self, engine, **kw):
        super(MongoDBStore, self).__init__(engine, **kw)
        spliturl = urlsplit(engine)
        _, dbpath, self._colpath = spliturl.path.split('/')
        self._conn = Connection(host=spliturl.hostname, port=spliturl.port)
        self._db = getattr(self._conn, dbpath)
        self._store = getattr(self._db, self._colpath)
        self._store.ensure_index('key', unique=True)

    def __getitem__(self, key):
        try:
            return self.loads(self._store.find_one(dict(key=key))['value'])
        except TypeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._store.save(dict(key=key, value=Binary(self.dumps(value))))

    def __delitem__(self, key):
        self._store.remove(dict(key=key))

    def __len__(self):
        return self._store.count()

    def __iter__(self):
        for key in self._store.find(
            dict(key={'$exists': True}), fields=['key'],
        ):
            yield key['key']

    def close(self):
        self._conn.close()

    def clear(self):
        self._store.drop()
        self._store = getattr(self._db, self._colpath)

    def items(self):
        for key in self._store.find({'key': {'$exists': True}}):
            yield (key['key'], key['value'])

    def values(self):
        loads = self.loads
        for value in self._store.distinct('key'):
            yield loads(value['value'])



# -*- coding: utf-8 -*-
'''
Redis-based object store

The shove URI for a redis-based store is:

redis://<host>:<port>/<db>
'''

try:
    import redis
except ImportError:
    raise ImportError('requires the redis library')

from shove._compat import urlsplit

from shove.store import BaseStore

__all__ = ['RedisStore']


class RedisStore(BaseStore):

    '''
    Redis based store.
    '''

    init = 'redis://'

    def __init__(self, engine, **kw):
        super(RedisStore, self).__init__(engine, **kw)
        spliturl = urlsplit(engine)
        self._db = spliturl.path.replace('/', '')
        self._hostname = spliturl.hostname
        self._port = spliturl.port
        self._store = redis.Redis(spliturl.hostname, spliturl.port, self._db)

    def __contains__(self, key):
        return self._store.exists(key)

    def __len__(self):
        return len(self._store.keys())

    def clear(self):
        self._store.flushall()

    def __iter__(self):
        return iter(self._store.keys())

    def keys(self):
        return self._store.keys()

    def setdefault(self, key, default=None):
        return self._store.getset(key, default)



# -*- coding: utf-8 -*-
'''
FTP-accessed stores

shove's URI for FTP accessed stores follows the standard form for FTP URLs
defined in RFC-1738:

ftp://<user>:<password>@<host>:<port>/<url-path>
'''

from ftplib import FTP, error_perm

from shove.store import BaseStore
from shove._compat import StringIO, urlsplit

__all__ = ['FTPStore']


class FTPStore(BaseStore):

    def __init__(self, engine, **kw):
        super(FTPStore, self).__init__(engine, **kw)
        user = kw.get('user', 'anonymous')
        password = kw.get('password', '')
        spliturl = urlsplit(engine)
        # Set URL, path, and strip 'ftp://' off
        base, path = spliturl[1], spliturl[2] + '/'
        if '@' in base:
            auth, base = base.split('@')
            user, password = auth.split(':')
        self._store = FTP(base, user, password)
        # Change to remote path if it exits
        try:
            self._store.cwd(path)
        except error_perm:
            self._makedir(path)
        self._base, self._user, self._password = base, user, password
        self._updated, self ._keys = True, None

    def __getitem__(self, key):
        try:
            local = StringIO()
            # Download item
            self._store.retrbinary('RETR %s' % key, local.write)
            self._updated = False
            return self.loads(local.getvalue())
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        local = StringIO(self.dumps(value))
        self._store.storbinary('STOR %s' % key, local)
        self._updated = True

    def __delitem__(self, key):
        try:
            self._store.delete(key)
            self._updated = True
        except:
            raise KeyError(key)

    def _makedir(self, path):
        # makes remote paths on an FTP server
        paths = list(reversed([i for i in path.split('/') if i != '']))
        store = self._store
        while paths:
            tpath = paths.pop()
            store.mkd(tpath)
            store.cwd(tpath)

    def keys(self):
        '''Returns a list of keys in a store.'''
        if self._updated or self._keys is None:
            rlist, nlist = list(), list()
            nappend = nlist.append
            # remote directory listing
            self._store.retrlines('LIST -a', rlist.append)
            for rlisting in rlist:
                # split remote file based on whitespace
                rfile = rlisting.split()
                # append tuple of remote item type & name
                if rfile[-1] not in ('.', '..') and rfile[0].startswith('-'):
                    nappend(rfile[-1])
            self._keys = nlist
        return self._keys



# -*- coding: utf-8 -*-
'''
Database object cache.

The shove URI used for database object caches is the format used by
SQLAlchemy:

<driver>://<username>:<password>@<host>:<port>/<database>

<driver> is the database engine. The engines currently supported SQLAlchemy are
sqlite, mysql, postgres, oracle, mssql, and firebird.
<username> is the database account user name
<password> is the database accound password
<host> is the database location
<port> is the database port
<database> is the name of the specific database

For more information on specific databases see:

http://www.sqlalchemy.org/docs/dbengine.myt#dbengine_supported
'''

import time
from random import sample
from datetime import datetime

from stuf.six import native
try:
    from sqlalchemy import LargeBinary as Binary
except ImportError:
    from sqlalchemy import Binary
try:
    from sqlalchemy import (
        MetaData, Table, Column, String, DateTime, select, update, insert,
        delete,
    )
except ImportError:
    raise ImportError('requires SQLAlchemy >= 0.4')

from shove.base import Base

__all__ = ['DBCache']


class DBCache(Base):

    '''
    Database cache backend.
    '''

    def __init__(self, engine, **kw):
        super(DBCache, self).__init__(engine, **kw)
        # make cache table
        self._store = Table(
            # get table name
            kw.get('tablename', 'cache'),
            # bind metadata
            MetaData(engine),
            Column('key', String(60), primary_key=True, nullable=False),
            Column('value', Binary, nullable=False),
            Column('expires', DateTime, nullable=False),
        )
        # create cache table if it does not exist
        if not self._store.exists():
            self._store.create()
        # set maximum entries
        self._max_entries = kw.get('max_entries', 300)
        # maximum number of entries to cull per call if cache is full
        self._maxcull = kw.get('maxcull', 10)
        # set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        row = select(
             [self._store.c.value, self._store.c.expires],
            self._store.c.key == key
        ).execute().fetchone()
        if row is not None:
            # remove if item expired
            if row.expires < datetime.now().replace(microsecond=0):
                del self[key]
                raise KeyError(key)
            return self.loads(native(row.value))
        raise KeyError(key)

    def __setitem__(self, key, value):
        value = self.dumps(value)
        # cull if too many items
        if len(self) >= self._max_entries:
            self._cull()
        # generate expiration time
        expires = datetime.fromtimestamp(
            time.time() + self.timeout
        ).replace(microsecond=0)
        # update database if key already present
        if key in self:
            cache = self._store
            update(
                cache,
                cache.c.key == key,
                dict(value=value, expires=expires),
            ).execute()
        # insert new key if key not present
        else:
            insert(
                self._store, dict(key=key, value=value, expires=expires)
            ).execute()

    def __delitem__(self, key):
        self._store.delete(self._store.c.key == key).execute()

    def __iter__(self):
        for item in select([self._store.c.key]).execute().fetchall():
            yield item[0]

    def __len__(self):
        return self._store.count().execute().fetchone()[0]

    def _cull(self):
        # remove items in cache to make more room
        cache = self._store
        # remove items that have timed out
        delete(
            cache, cache.c.expires < datetime.now().replace(microsecond=0),
        ).execute()
        # remove any items over the maximum allowed number in the cache
        length = len(self)
        if length >= self._max_entries:
            cull = length if length < self._maxcull else self._maxcull
            # get list of keys
            keys = list(i[0] for i in select(
                [cache.c.key], limit=cull * 2
            ).execute().fetchall())
            # delete keys at random
            delete(cache, cache.c.key.in_(sample(keys, cull))).execute()



# -*- coding: utf-8 -*-
'''shove loadpoints.'''

from stuf.six import strings
from stuf.utils import lazyimport
from pkg_resources import iter_entry_points

stores = dict(
    (_store.name, _store) for _store in iter_entry_points('shove.stores')
)
caches = dict(
    (_cache.name, _cache) for _cache in iter_entry_points('shove.caches')
)


def cache_backend(uri, **kw):
    '''
    Loads the right cache backend based on a URI.

    :argument uri: instance or name :class:`str`
    '''
    if isinstance(uri, strings):
        mod = caches[uri.split('://', 1)[0]]
        # load module if setuptools not present
        if isinstance(mod, strings):
            # split classname from dot path
            module, klass = mod.split(':')
            # load module
            mod = lazyimport(module, klass)
        # load appropriate class from setuptools entry point
        else:
            mod = mod.load()
        # return instance
        return mod(uri, **kw)
    # no-op for existing instances
    return uri


def store_backend(uri, **kw):
    '''
    Loads the right store backend based on a URI.

    :argument uri: instance or name :class:`str`
    '''
    if isinstance(uri, strings):
        mod = stores[uri.split('://', 1)[0]]
        # load module if setuptools not present
        if isinstance(mod, strings):
            # isolate classname from dot path
            module, klass = mod.split(':')
            # load module
            mod = lazyimport(module, klass)
        # load appropriate class from setuptools entry point
        else:
            mod = mod.load()
        # return instance
        return mod(uri, **kw)
    # no-op for existing instances
    return uri



# -*- coding: utf-8 -*-
'''
shove cache core
'''

import random
from time import time
from copy import deepcopy
from collections import deque
from threading import Condition

from shove._compat import synchronized
from shove.base import Mapping, FileBase

__all__ = [
    'FileCache', 'FileLRUCache', 'MemoryCache', 'MemoryLRUCache',
    'SimpleLRUCache', 'SimpleCache',
]


class BaseCache(object):

    def __init__(self, engine, **kw):
        super(BaseCache, self).__init__(engine, **kw)
        # get random seed
        random.seed()
        # set maximum number of items to cull if over max
        self._maxcull = kw.get('maxcull', 10)
        # set max entries
        self._max_entries = kw.get('max_entries', 300)
        # set timeout
        self.timeout = kw.get('timeout', 300)

    def __getitem__(self, key):
        exp, value = super(BaseCache, self).__getitem__(key)
        # delete if item timed out.
        if exp < time():
            super(BaseCache, self).__delitem__(key)
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        # cull values if over max number of entries
        if len(self) >= self._max_entries:
            self._cull()
        # set expiration time and value
        exp = time() + self.timeout
        super(BaseCache, self).__setitem__(key, (exp, value))

    def _cull(self):
        # remove items in cache to make room
        maxcull = self._maxcull
        # cull number of items allowed (set by self._maxcull)
        for num, key in enumerate(self):
            # remove only maximum # of items allowed by maxcull
            if num <= maxcull:
                # remove items if expired
                try:
                    self[key]
                except KeyError:
                    num += 1
            else:
                break
        choice = random.choice
        keys = list(self)
        max_entries = self._max_entries
        # remove any additional items up to max # of items allowed by maxcull
        while len(self) >= max_entries and num <= maxcull:
            # cull remainder of allowed quota at random
            del self[choice(keys)]
            num += 1


class SimpleCache(BaseCache, Mapping):

    '''
    Single-process in-memory cache.

    The shove URI for a simple cache is:

    simple://
    '''

    def __init__(self, engine, **kw):
        super(SimpleCache, self).__init__(engine, **kw)
        self._store = dict()


class MemoryCache(SimpleCache):

    '''
    Thread-safe in-memory cache.

    The shove URI for a memory cache is:

    memory://
    '''

    def __init__(self, engine, **kw):
        super(MemoryCache, self).__init__(engine, **kw)
        self._lock = Condition()

    @synchronized
    def __getitem__(self, key):
        return deepcopy(super(MemoryCache, self).__getitem__(key))

    __setitem__ = synchronized(SimpleCache.__setitem__)
    __delitem__ = synchronized(SimpleCache.__delitem__)


class FileCache(BaseCache, FileBase):

    '''
    File-based cache

    shove's URI for file caches follows the form:

    file://<path>

    Where the path is a URI path to a directory on a local filesystem.
    Alternatively, a native pathname to the directory can be passed as the
    'engine' argument.
    '''

    init = 'file://'

    def __init__(self, engine, **kw):
        super(FileCache, self).__init__(engine, **kw)

    def __getitem__(self, key):
        try:
            exp, value = super(FileCache, self).__getitem__(key)
            # remove item if time has expired.
            if exp < time():
                del self[key]
                raise KeyError(key)
            return value
        except:
            raise KeyError(key)

    def __setitem__(self, key, value):
        if len(self) >= self._max_entries:
            self._cull()
        super(FileCache, self).__setitem__(
            key, (time() + self.timeout, value)
        )


class BaseLRUCache(BaseCache):

    def __init__(self, engine, **kw):
        super(BaseLRUCache, self).__init__(engine, **kw)
        self._max_entries = kw.get('max_entries', 300)
        self._hits = 0
        self._misses = 0
        self._queue = deque()
        self._refcount = dict()

    def __getitem__(self, key):
        try:
            value = super(BaseLRUCache, self).__getitem__(key)
            self._hits += 1
        except KeyError:
            self._misses += 1
            raise
        self._housekeep(key)
        return value

    def __setitem__(self, key, value):
        super(BaseLRUCache, self).__setitem__(key, value)
        self._housekeep(key)
        if len(self) > self._max_entries:
            queue = self._queue
            store = self
            max_entries = self._max_entries
            refcount = self._refcount
            delitem = super(BaseLRUCache, self).__delitem__
            while len(store) > max_entries:
                k = queue.popleft()
                refcount[k] -= 1
                if not refcount[k]:
                    delitem(k)
                    del refcount[k]

    def _housekeep(self, key):
        self._queue.append(key)
        self._refcount[key] = self._refcount.get(key, 0) + 1
        if len(self._queue) > self._max_entries * 4:
            queue = self._queue
            refcount = self._refcount
            for _ in [None] * len(queue):
                k = queue.popleft()
                if refcount[k] == 1:
                    queue.append(k)
                else:
                    refcount[k] -= 1


class SimpleLRUCache(BaseLRUCache, Mapping):

    '''
    Single-process in-memory LRU cache that purges based on least recently
    used item.

    The shove URI for a simple cache is:

    simplelru://
    '''

    def __init__(self, engine, **kw):
        super(SimpleLRUCache, self).__init__(engine, **kw)
        self._store = dict()


class MemoryLRUCache(SimpleLRUCache):

    '''
    Thread-safe in-memory cache using LRU.

    The shove URI for a memory cache is:

    memlru://
    '''

    def __init__(self, engine, **kw):
        super(MemoryLRUCache, self).__init__(engine, **kw)
        self._lock = Condition()

    @synchronized
    def __getitem__(self, key):
        return deepcopy(super(MemoryLRUCache, self).__getitem__(key))

    __setitem__ = synchronized(SimpleLRUCache.__setitem__)
    __delitem__ = synchronized(SimpleLRUCache.__delitem__)


class FileLRUCache(BaseLRUCache, FileBase):

    '''
    File-based LRU cache

    shove's URI for file caches follows the form:

    filelru://<path>

    Where the path is a URI path to a directory on a local filesystem.
    Alternatively, a native pathname to the directory can be passed as the
    'engine' argument.
    '''

    init = 'filelru://'

    def __init__(self, engine, **kw):
        super(FileLRUCache, self).__init__(engine, **kw)



# -*- coding: utf-8 -*-
'''shove compatibility for different python versions'''

try:
    from urlparse import urlsplit
    from urllib import quote_plus, unquote_plus
except ImportError:
    from urllib.parse import urlsplit, quote_plus, unquote_plus
try:
    from urllib import url2pathname
except ImportError:
    from urllib.request import url2pathname
try:
    import anydbm
except ImportError:
    import dbm as anydbm

from stuf.six import PY3
from stuf.six.moves import StringIO  # @UnresolvedImport


def synchronized(func):
    '''
    Decorator to lock and unlock a method (Phillip J. Eby).

    @param func Method to decorate
    '''
    def wrapper(self, *__args, **__kw):
        self._lock.acquire()
        try:
            return func(self, *__args, **__kw)
        finally:
            self._lock.release()
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper


def openit(path, mode, encoding='utf-8'):
    return open(path, mode, encoding=encoding) if PY3 else open(path, mode)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
setup for shove.
'''

from setuptools import setup, find_packages

setup(
    name='shove',
    version='0.5.0',
    description='''Common object storage frontend''',
    long_description=open('README.rst').read(),
    author='L. C. Rees',
    author_email='lcrees@gmail.com',
    url='https://bitbucket.org/lcrees/shove/',
    license='BSD',
    packages=find_packages(),
    test_suite='shove.tests',
    install_requires=['futures', 'stuf'],
    zip_safe=False,
    keywords='object storage persistence database shelve',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Database :: Front-Ends',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points='''
    [shove.stores]
    bsddb=shove.stores.bsdb:BSDBStore
    cassandra=shove.stores.cassandra:CassandraStore
    dbm=shove.store:DBMStore
    durus=shove.stores.durusdb:DurusStore
    file=shove.store:FileStore
    firebird=shove.stores.db:DBStore
    ftp=shove.stores.ftp:FTPStore
    hdf5=shove.stores.hdf5:HDF5Store
    leveldb=shove.stores.leveldbstore:LevelDBStore
    memory=shove.store:MemoryStore
    mongodb=shove.stores.mongodb:MongoDBStore
    mssql=shove.store.db:DBStore
    mysql=shove.stores.db:DBStore
    oracle=shove.stores.db:DBStore
    postgres=shove.store.db:DBStore
    redis=shove.stores.redisdb:RedisStore
    s3=shove.stores.s3:S3Store
    simple=shove.store:SimpleStore
    sqlite=shove.stores.db:DBStore
    zodb=shove.stores.zodb:ZODBStore
    hg=shove.stores.hgstore:HgStore
    git=shove.stores.gitstore:GitStore
    [shove.caches]
    file=shove.cache:FileCache
    filelru=shove.cache:FileLRUCache
    firebird=shove.caches.db:DBCache
    memcache=shove.caches.memcached:MemCache
    memlru=shove.cache:MemoryLRUCache
    memory=shove.caches:MemoryCache
    mssql=shove.caches.db:DBCache
    mysql=shove.caches.db:DBCache
    oracle=shove.caches.db:DBCache
    postgres=shove.caches.db:DBCache
    redis=shove.caches.redisdb:RedisCache
    simple=shove.cache:SimpleCache
    simplelru=shove.cache:SimpleLRUCache
    sqlite=shove.caches.db:DBCache
    ''',
)



# -*- coding: utf-8 -*-
'''shove core.'''

from functools import partial
from operator import methodcaller
from collections import MutableMapping

from stuf.iterable import exhaustcall
from concurrent.futures import ThreadPoolExecutor

from shove._imports import cache_backend, store_backend

__all__ = ('Shove', 'MultiShove')


class Shove(MutableMapping):

    '''
    Common object frontend class.
    '''

    def __init__(self, store='simple://', cache='simple://', **kw):
        super(Shove, self).__init__()
        # load store backend
        self._store = store_backend(store, **kw)
        # load cache backend
        self._cache = cache_backend(cache, **kw)
        # buffer for lazier writing
        self._buffer = dict()
        # setting for syncing frequency
        self._sync = kw.get('sync', 2)

    def __getitem__(self, key):
        '''
        Gets a item from shove.
        '''
        try:
            return self._cache[key]
        except KeyError:
            # synchronize cache with store
            self.sync()
            self._cache[key] = value = self._store[key]
            return value

    def __setitem__(self, key, value):
        '''
        Sets an item in shove.
        '''
        self._cache[key] = self._buffer[key] = value
        # when buffer reaches self._limit, write buffer to store
        if len(self._buffer) >= self._sync:
            self.sync()

    def __delitem__(self, key):
        '''
        Deletes an item from shove.
        '''
        self.sync()
        try:
            del self._cache[key]
        except KeyError:
            pass
        del self._store[key]

    def __len__(self):
        return len(self._store)

    def __iter__(self):
        self.sync()
        return self._store.__iter__()

    def close(self):
        '''
        Finalizes and closes shove.
        '''
        # if close has been called, pass
        if self._store is not None:
            try:
                self.sync()
            except AttributeError:
                pass
            self._store.close()
        self._store = self._cache = self._buffer = None

    def sync(self):
        '''
        Writes buffer to store.
        '''
        self._store.update(self._buffer)
        self._buffer.clear()


class MultiShove(MutableMapping):

    '''
    Common frontend to multiple object stores.
    '''

    def __init__(self, *stores, **kw):
        # init superclass with first store
        super(MultiShove, self).__init__()
        if not stores:
            stores = ('simple://',)
        # load stores
        self._stores = list(store_backend(i, **kw) for i in stores)
        # load cache
        self._cache = cache_backend(kw.get('cache', 'simple://'), **kw)
        # buffer for lazy writing
        self._buffer = dict()
        # setting for syncing frequency
        self._sync = kw.get('sync', 2)

    def __getitem__(self, key):
        '''
        Get a item from shove.
        '''
        try:
            return self._cache[key]
        except KeyError:
            # synchronize cache and store
            self.sync()
            # get value from first store
            self._cache[key] = value = self._stores[0][key]
            return value

    def __setitem__(self, key, value):
        '''
        Sets an item in shove.
        '''
        self._cache[key] = self._buffer[key] = value
        # when the buffer reaches self._limit, writes the buffer to the store
        if len(self._buffer) >= self._sync:
            self.sync()

    def __delitem__(self, key):
        '''
        Deletes an item from multiple stores.
        '''
        try:
            self.sync()
        except AttributeError:
            pass
        exhaustcall(methodcaller('__delitem__', key), self._stores)
        try:
            del self._cache[key]
        except KeyError:
            pass

    def __len__(self):
        return len(self._stores[0])

    def __iter__(self):
        self.sync()
        return self._stores[0].__iter__()

    def close(self):
        '''
        Finalizes and closes shove stores.
        '''
        # if close has been called, pass
        stores = self._stores
        if self._stores is not None:
            self.sync()
            # close stores
            for idx, store in enumerate(stores):
                store.close()
                stores[idx] = None
        self._cache = self._buffer = self._stores = None

    def sync(self):
        '''
        Writes buffer to stores.
        '''
        exhaustcall(methodcaller('update', self._buffer), self._stores)
        self._buffer.clear()


class ThreadShove(MultiShove):

    '''
    Common frontend that syncs multiple object stores with multiple threads.
    '''

    def __init__(self, *stores, **kw):
        # init superclass with first store
        super(ThreadShove, self).__init__(*stores, **kw)
        self._maxworkers = kw.get('max_workers', 2)

    def __delitem__(self, key):
        try:
            self.sync()
        except AttributeError:
            pass
        with ThreadPoolExecutor(max_workers=self._maxworkers) as executor:
            method = partial(
                executor.submit, methodcaller('__delitem__', key)
            )
            exhaustcall(method, self._stores)
        try:
            del self._cache[key]
        except KeyError:
            pass

    def sync(self):
        '''
        Writes buffer to store.
        '''
        with ThreadPoolExecutor(max_workers=self._maxworkers) as executor:
            method = partial(
                executor.submit, methodcaller('update', self._buffer),
            )
            exhaustcall(method, self._stores)
        self._buffer.clear()



# -*- coding: utf-8 -*-
'''shove cache tests'''

from stuf.six import PY3, unittest


def setUpModule():
    import os
    from tempfile import mkdtemp
    TMP = mkdtemp()
    os.chdir(TMP)


class NoTimeout(object):

    def setUp(self):
        self.cache = self._makeone(self.initstring)

    def tearDown(self):
        self.cache = None

    def test_getitem(self):
        self.cache['test'] = 'test'
        self.assertEqual(self.cache['test'], 'test')

    def test_setitem(self):
        self.cache['test'] = 'test'
        self.assertEqual(self.cache['test'], 'test')

    def test_delitem(self):
        self.cache['test'] = 'test'
        del self.cache['test']
        self.assertEqual('test' in self.cache, False)


class Cache(NoTimeout):

    def test_timeout(self):
        import time
        cache = self._makeone(self.initstring, timeout=1)
        cache['test'] = 'test'
        time.sleep(3)
        def tmp(): #@IgnorePep8
            cache['test']
        self.assertRaises(KeyError, tmp)


class CacheCull(Cache):

    def test_cull(self):
        cache = self._makeone(self.initstring, max_entries=1)
        cache['test'] = 'test'
        cache['test2'] = 'test'
        cache['test2'] = 'test'
        self.assertEquals(len(cache), 1)


class TestSimpleCache(CacheCull, unittest.TestCase):

    initstring = 'simple://'

    @property
    def _makeone(self):
        from shove.cache import SimpleCache
        return SimpleCache


class TestSimpleLRUCache(NoTimeout, unittest.TestCase):

    initstring = 'simplelru://'

    @property
    def _makeone(self):
        from shove.cache import SimpleLRUCache
        return SimpleLRUCache


class TestMemoryCache(CacheCull, unittest.TestCase):

    initstring = 'memory://'

    @property
    def _makeone(self):
        from shove.cache import MemoryCache
        return MemoryCache


class TestMemoryLRUCache(NoTimeout, unittest.TestCase):

    initstring = 'memlru://'

    @property
    def _makeone(self):
        from shove.cache import MemoryLRUCache
        return MemoryLRUCache


class TestFileCache(CacheCull, unittest.TestCase):

    initstring = 'file://test'

    @property
    def _makeone(self):
        from shove.cache import FileCache
        return FileCache

    def tearDown(self):
        import shutil
        self.cache = None
        shutil.rmtree('test')


class TestFileLRUCache(NoTimeout, unittest.TestCase):

    initstring = 'filelru://test2'

    @property
    def _makeone(self):
        from shove.cache import FileLRUCache
        return FileLRUCache

    def tearDown(self):
        import shutil
        self.cache = None
        shutil.rmtree('test2')


class TestDBCache(CacheCull, unittest.TestCase):

    initstring = 'sqlite:///'

    @property
    def _makeone(self):
        from shove.caches.db import DBCache
        return DBCache


if not PY3:
    class TestMemcache(Cache, unittest.TestCase):

        initstring = 'memcache://localhost:11211'

        @classmethod
        def setUpClass(cls):
            from fabric.api import local
            local('memcached -d')

        @property
        def _makeone(self):
            from shove.caches.memcached import MemCache
            return MemCache

        @classmethod
        def tearDownClass(cls):
            from fabric.api import local
            local('killall memcached')

    class TestRedisCache(Cache, unittest.TestCase):

        initstring = 'redis://localhost:6379/0'

        @classmethod
        def setUpClass(cls):
            from fabric.api import local
            local('redis-server &')

        @property
        def _makeone(self):
            from shove.caches.redisdb import RedisCache
            return RedisCache

        @classmethod
        def tearDownClass(cls):
            from fabric.api import local
            local('killall redis-server')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''
Berkeley Source Database Store.

shove's URI for BSDDB stores follows the form:

bsddb://<path>

Where the path is a URL path to a Berkeley database. Alternatively, the native
pathname to a Berkeley database can be passed as the 'engine' parameter.
'''

from threading import Condition

from stuf.six import b
try:
    from bsddb import hashopen
except ImportError:
    try:
        from bsddb3 import hashopen
    except ImportError:
        raise ImportError('requires bsddb library')

from shove.store import SyncStore
from shove._compat import synchronized

__all__ = ['BSDBStore']


class BSDBStore(SyncStore):

    '''
    Berkeley Source Database store.
    '''

    init = 'bsddb://'

    def __init__(self, engine, **kw):
        super(BSDBStore, self).__init__(engine, **kw)
        self._store = hashopen(self._engine)
        self._lock = Condition()
        self.sync = self._store.sync

    @synchronized
    def __getitem__(self, key):
        return self.loads(self._store[key])

    @synchronized
    def __setitem__(self, key, value):
        self._store[b(key)] = self.dumps(value)
        self.sync()

    @synchronized
    def __delitem__(self, key):
        del self._store[b(key)]
        self.sync()



# -*- coding: utf-8 -*-
'''
Cassandra-based object store

The shove URI for a cassandra-based store is:

cassandra://<host>:<port>/<keyspace>/<columnFamily>
'''

from stuf.six import native

try:
    import pycassa
except ImportError:
    raise ImportError('requires pycassa library')

from shove._compat import urlsplit
from shove.store import SimpleStore

__all__ = ['CassandraStore']


class CassandraStore(SimpleStore):

    '''Cassandra based store'''

    init = 'cassandra://'

    def __init__(self, engine, **kw):
        super(CassandraStore, self).__init__(engine, **kw)
        spliturl = urlsplit(engine)
        _, keyspace, column_family = spliturl.path.split('/')
        try:
            self._store = pycassa.ColumnFamily(
                pycassa.ConnectionPool(keyspace, [spliturl.hostname]),
                column_family,
            )
        except pycassa.InvalidRequestException:
            from pycassa.system_manager import SystemManager  # @UnresolvedImport @IgnorePep8
            system_manager = SystemManager(spliturl[1])
            system_manager.create_keyspace(
                keyspace,
                pycassa.system_manager.SIMPLE_STRATEGY,
                dict(replication_factor=native(kw.get('replication', 1))),
            )
            system_manager.create_column_family(keyspace, column_family)
            self._store = pycassa.ColumnFamily(
                pycassa.ConnectionPool(keyspace, [spliturl.netloc]),
                column_family,
            )

    def __getitem__(self, key):
        try:
            item = self._store.get(key).get(key)
            if item is not None:
                return self.loads(item)
            raise KeyError(key)
        except pycassa.NotFoundException:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._store.insert(key, dict(key=self.dumps(value)))

    def __delitem__(self, key):
        # beware eventual consistency
        try:
            self._store.remove(key)
        except pycassa.NotFoundException:
            raise KeyError(key)

    def __len__(self):
        return len(list(self._store.get_range()))

    def __iter__(self):
        for item in self._store.get_range():
            yield item[0]

    def clear(self):
        # beware eventual consistency
        self._store.truncate()
